import os, shutil, subprocess, datetime, json, requests


def detect_code_root():
    """
    Expected directory structure:
    my-code/
    my-code/.smc_code_root     -- indicates that this folder and all its subfolders necessary to run your code.
    my-code/necessary-thing1/
    my-code/necessary-thing2/
    my-code/my-script/this-does-the-work.py     -- imports thing1, thing2, does the work.

    This function returns path to 'my-code', a path to be uploaded to a remote machine or packed into a zip.
    It works by looking at the current working directory and up, looing for a file named '.smc_code_root'.
    """
    start_dir = os.getcwd()
    p = start_dir
    while 1:
        if os.path.exists(p + "/.smc_code_root.sh"):
            break
        if p == os.path.dirname(p):
            print("Cannot find code root, searched the current directory '%s' and up." % start_dir)
            print("Please create a file '.smc_code_root.sh' in the directory you want to upload to your VM, for example:")
            print(f"touch {start_dir}/.smc_code_root.sh")
            print("You can leave this file empty or put initialization commands here, such as \"pip install -r requirements.txt\" or \"echo 'export VAR=value' >> ~/.profile\"")
            quit(0)
        p = os.path.dirname(p)
    p += "/"  # that makes rsync happy
    print("Code root detected at: %s" % p)
    return p


def code_to_zip():
    fn = "codezip_" + datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
    assert shutil.which("7zr") is not None, "ubuntu: apt-get install p7zip-full\nmac: brew install p7zip"
    root = detect_code_root()
    path_from_home = root.replace(os.path.expanduser('~/'), '', 1)
    # If you use several code roots, path_from_home helps rsync destination to be unique.
    tmp = os.path.join("/tmp/smc-temp/", path_from_home)
    os.makedirs(tmp, exist_ok=True)
    cmd = ["rsync", "-rplu", "--delete", ".", tmp, "--filter=:- .gitignore", "--exclude=.git"]
    print(" ".join(cmd))
    subprocess.check_call(cmd, cwd=root, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    cmd = ["7za", "-bso0", "-y", "a", f"/tmp/smc-temp/{fn}.7z", "."]
    # -bs{o|e|p}{0|1|2} : set output stream for output/error/progress line
    print(" ".join(cmd))
    subprocess.check_call(cmd, cwd=tmp, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    return f"/tmp/smc-temp/{fn}.7z"



import os, sys, json, time, termcolor
from smallcloud import config, call_api, ssh_commands

# This file runs on "s" in command line.


def printhl(s):
    print(termcolor.colored(s, attrs=["bold"]))


def print_help():
    print("This is a command line tool to use Small Magellanic Cloud AI Ltd services.")
    print("Homepage for this tool:")
    print("    https://github.com/smallcloudai/smallcloud")
    print("Commands:")
    printhl("s free")
    print("      Print number of free GPUs, works without login.")
    printhl("s prices")
    print("      Print prices, works without login.")
    printhl("s login")
    print("      Interactive login using your web browser.")
    printhl("s list")
    print("      Prints your jobs, working and finished.")
    printhl("s reserve <gpu_type> <gpu_count> <job_name>")
    print("      Reserve GPUs, start the job. Valid gpu_count values are 1, 2, 4, 8, 16, 32, 64.")
    print("      Starting from 16, multiple VMs will be launched.")
    print("      If the job cannot start immediately, it will be queued.")
    printhl("s delete <job_name>")
    print("      Delete the job. Use \"experiment05*\" syntax to delete several jobs.")
    printhl("s ssh <job_name> [<any-ssh-args>]")
    print("      SSH into the job. By default the user is \"user\". You can use \"otheruser@jobname\" syntax if you created more users.")
    printhl("s scp <local_file> <job_name>:<remote_file> [<any-scp-args>]")
    print("      Copy a file.")
    printhl("s upload-code <job_name>")
    print("      Upload your source code using rsync.")
    print("      Use \"experiment05*\" syntax to upload to several jobs.")
    print("      Remote destination is hardcoded as \"/home/user/code/\".")
    print("      Add --init option to run \"/home/user/code/.smc_code_root.sh\" after the upload is completed.")
    printhl("s ssh-keygen")
    print("      Generate a dedicated SSH keypair and upload the public part.")
    printhl("s ssh-key-upload")
    print("      If you prefer, you can upload this computer's public key or any other key.")
    printhl("s billing")
    printhl("s billing-detailed")
    printhl("s money")
    print("      CLI analogs of webpages to monitor your balance and billing.")


def command_login(*args):
    assert len(args) <= 1
    print("Please open this link in your browser:\n")
    print(termcolor.colored(config.v1_url.replace("/v1/", "/cli-login"), attrs=["bold"]))
    ticket = input("\nand copy-paste a response here: ")
    resp = call_api.fetch_json(config.v1_url + "cli-login-response", get_params={"ticket": ticket})
    os.makedirs(config.config_dir, exist_ok=True)
    with open(config.config_file, "w") as f:
        f.write(json.dumps({
            "account_name": resp["account_name"],
            "expires_ts": (resp["expires_ts"] if "expires_ts" in resp else 365*24*60*60 + time.time()),
            "secret_api_key": resp["secret_api_key"],
            }, indent=4))
    os.chmod(config.config_file, 0o600)
    print("Login successful: %s" % resp["account_name"])
    print("Account name and the Secret API Key were stored in %s" % config.config_file)
    print("Try this:")
    print(termcolor.colored("s list", attrs=["bold"]))
    print(termcolor.colored("s free", attrs=["bold"]))
    print(termcolor.colored("s reserve a5000 4 myexperiment00", attrs=["bold"]))


def command_logout():
    if not config.username:
        print("You are not logged in")
        return
    os.remove(config.config_file)
    print("Logged out")


def command_free():
    free_json = call_api.fetch_json(config.v1_url + "free", headers=config.account_and_secret_key_headers())
    call_api.print_table(free_json)


def command_reserve(*args):
    config.make_sure_have_login()
    # The only nontrivial command with options at this point:
    import argparse
    parser = argparse.ArgumentParser(description="Schedule your GPU task.")
    subparsers = parser.add_subparsers()
    parser_reserve = subparsers.add_parser("reserve")
    parser_reserve.add_argument("gpu_type", help="GPU to reserve")
    parser_reserve.add_argument("count", type=int, help="Number of GPUs")
    parser_reserve.add_argument("job_name", help="Name of the experiment")
    parser_reserve.add_argument("--os", help="Operating system")
    parser_reserve.add_argument("--force-node", help=argparse.SUPPRESS)
    args = parser.parse_args(("reserve",) + args)
    gpus_min = args.count
    post_json = {
        "task_name": args.job_name,
        "gpu_type": args.gpu_type,
        "gpus_min": int(gpus_min),
        "force_node": args.force_node,
        }
    if args.os:
        post_json["tenant_image"] = args.os
    ret_json = call_api.fetch_json(config.v1_url + "reserve", post_json, headers=config.account_and_secret_key_headers())
    call_api.pretty_print_response(ret_json)


def command_jobs(*args):
    config.make_sure_have_login()
    resp = call_api.fetch_json(config.v1_url + "jobs", headers=config.account_and_secret_key_headers())
    day_ago = time.time() - 24*3600
    if resp == []:
        print("There are no jobs yet. You can start one using:\n" + termcolor.colored("s reserve a5000 4 myexperiment00-seed0", attrs=["bold"]))
        return
    import argparse
    parser = argparse.ArgumentParser(description="List your jobs.")
    parser.add_argument("--all", action="store_true", help="Show all jobs, not only those that are finished less than a day ago.")
    args = parser.parse_args(args)
    if args.all:
        finished_less_than_day_ago = resp
    else:
        finished_less_than_day_ago = [x for x in resp if x["ts_finished"] == 0 or x["ts_finished"] > day_ago]
    hidden = len(resp) - len(finished_less_than_day_ago)
    if hidden:
        print(termcolor.colored("Finished more than a day ago: %i" % hidden, "white"))
    call_api.print_table(finished_less_than_day_ago, ["cluster_name", "tenant_image", "ts_placed", "gpu_type", "gpus_min", "gpus_max", "gpus_incr", "nice", "ed25519"])


def command_delete(*task_names):
    config.make_sure_have_login()
    for tname in task_names:
        resp = call_api.fetch_json(config.v1_url + "delete", get_params={"task_name": tname}, headers=config.account_and_secret_key_headers())
        call_api.pretty_print_response(resp)


def command_nodes():
    nodes_json = call_api.fetch_json(config.v1_url + "nodes", headers=config.account_and_secret_key_headers())
    call_api.print_table(nodes_json)


def command_promo(*args):
    if len(args) == 0:
        print("This command applies a promo code (might add money to your account).")
        return
    assert len(args) == 1
    resp = call_api.fetch_json(config.v1_url + "apply-promo", get_params={"code": args[0]}, headers=config.account_and_secret_key_headers())
    call_api.pretty_print_response(resp)


def command_billing(subcmd):
    resp = call_api.fetch_json(config.v1_url + subcmd, headers=config.account_and_secret_key_headers())
    if subcmd == "money":
        print(json.dumps(resp, indent=2))
    else:
        call_api.print_table(resp)


def command_prices():
    resp = call_api.fetch_json(config.v1_url + "prices")
    print(resp)


def cli_command(command, *args):
    if command == "free":
        command_free()

    elif command == "login":
        command_login(*args)

    elif command == "logout":
        command_logout()

    elif command == "reserve":
        command_reserve(*args)

    elif command in ["list", "jobs"]:
        command_jobs(*args)

    elif command in ["delete", "remove"]:
        command_delete(*args)

    elif command == "nodes":
        command_nodes()

    elif command in ["upload-code", "code-upload"]:
        ssh_commands.command_upload_code(*args)

    elif command == "ssh":
        ssh_commands.command_ssh(*args)

    elif command == "scp":
        ssh_commands.command_scp(*args)

    elif command == "tail":
        ssh_commands.command_ssh(*args + ("tail -n 1000 -f output.log",))

    elif command == "ssh-keygen":
        ssh_commands.command_ssh_keygen(*args)

    elif command == "ssh-key-upload":
        ssh_commands.command_ssh_key_upload(*args)

    elif command == "promo":
        command_promo(*args)

    elif command == "billing":
        command_billing("billing-short")

    elif command == "billing-detailed":
        command_billing("billing-detailed")

    elif command in ["$", "money", "dollars"]:
        command_billing("money")

    elif command == "prices":
        command_prices()

    else:
        print_help()
        print("Unknown command:", command)
        quit(1)


def main():
    if "--json" in sys.argv:
        call_api.global_option_json = True
        sys.argv.remove("--json")
    if len(sys.argv) < 2:
        print_help()
        quit(0)
    config.read_config_file()
    cli_command(sys.argv[1], *sys.argv[2:])


if __name__=="__main__":
    main()



import os, json, time


v1_url = "https://www.smallcloud.ai/v1/"
if os.environ.get("staging"):
    v1_url = v1_url.replace("www", "staging")
if os.environ.get("local"):
    v1_url = v1_url.replace("www", "local")

config_dir = os.path.expanduser("~/.config/smallcloud.ai")
config_file = config_dir + "/cli_config"
ssh_rsa_id_file = config_dir + "/dedicated_ssh_rsa_id"
known_hosts_file = config_dir + "/known_hosts"
username = None
secret_api_key = None
already_running_in_cloud = os.path.exists("/etc/profile.d/50-smc.sh")


def read_config_file():
    if not os.path.exists(config_file):
        return
    try:
        with open(config_file, "r") as f:
            config = json.loads(f.read())
    except ValueError:
        print("Empty or invalid config file: %s (delete it to relogin)" % config_file)
        quit(1)
    global username, secret_api_key
    if "expires_ts" in config and time.time() > config["expires_ts"] > 0:
        print("Your login credentials are expired, please re-login.")
    else:
        username = config["account_name"]
        secret_api_key = config["secret_api_key"]


def account_and_secret_key_headers():
    if not username:
        return {}   # some commands work without login
    return {
        "Authorization": "Bearer " + secret_api_key,
    }


def make_sure_have_login():
    if not username:
        print("Please login to complete this operation, try \"s login\"")
        quit(1)



import sys, json, re, requests, time, datetime, termcolor, multiprocessing, copy, queue
from typing import Dict, Any, List, Optional, Set


url_base1 = "https://inference.smallcloud.ai/infengine-v1/"
url_base2 = "https://inference-backup.smallcloud.ai/infengine-v1/"
urls_to_try = [url_base1, url_base2]

urls_switch_n = 0
urls_switch_ts = time.time()


def log(*args):
    sys.stderr.write(" ".join([str(x) for x in args]) + "\n")
    sys.stderr.flush()


def url_get_the_best():
    global urls_switch_n, urls_switch_ts
    if time.time() > urls_switch_ts + 600:
        urls_switch_n = 0
    return urls_to_try[urls_switch_n]


def url_complain_doesnt_work():
    global urls_switch_n, urls_switch_ts
    urls_switch_n = (urls_switch_n + 1) % len(urls_to_try)
    urls_switch_ts = time.time()


def model_guid_allowed_characters(name):
    return re.sub(r"[^a-zA-Z0-9_]", "_", name)


def validate_description_dict(
    infeng_instance_guid: str,
    account: str,
    model: str,
    B: int,
    max_thinking_time: int,
    *,
    T: int = 0,               # deprecated
    encoding_name: str = "",  # deprecated
):
    return {
        "infmod_guid": model_guid_allowed_characters(infeng_instance_guid),
        "account": account,
        "model": model,
        "B": B,
        "engine_started_ts": int(time.time()),
        "ts_batch_started": 0,
        "ts_batch_finished": 0,
        "max_thinking_time": max_thinking_time,
    }


def completions_wait_batch(req_session, my_desc, verbose=False):
    resp = None
    json_resp = None
    for attempt in range(5):
        t0 = time.time()
        url = url_get_the_best() + "completions-wait-batch"
        try:
            resp = req_session.post(url, json=my_desc, timeout=15)
            json_resp = resp.json()
        except requests.exceptions.ReadTimeout as e:
            t1 = time.time()
            log("%s %0.1fms %s %s" % (datetime.datetime.now().strftime("%H:%M:%S.%f"), 1000*(t1 - t0), url, termcolor.colored("TIMEOUT", "green")))
            url_complain_doesnt_work()
            continue
        except Exception as e:
            log("%s fetch batch failed: %s %s\nServer response was: \"%s\"" % (url, str(type(e)), str(e), resp.text[:150] if resp else "no response"))
            # if resp is not None:
            #     log("server response text:\n%s" % (resp.text,))
            url_complain_doesnt_work()
            continue
        if resp and resp.status_code != 200:
            log("%s status_code %i %s" % (url, resp.status_code, resp.text))
            url_complain_doesnt_work()
            continue
        break
    if json_resp is None:
        return "ERROR", []
    t1 = time.time()
    hms = datetime.datetime.now().strftime("%H:%M:%S.%f")
    log("%s %0.1fms %s %s" % (hms, 1000*(t1 - t0), url, termcolor.colored(json_resp.get("retcode", "no retcode"), "green")))
    if verbose:
        log("%s %s" % (url, json.dumps(json_resp, indent=4)))
    return json_resp["retcode"], json_resp.get("batch", [])


def head_and_tail(base: str, modified: str):
    """
    Finds common head and tail of two strings.
    Returns tuple (head, tail) in chars.
    """
    head = 0
    tail = 0
    l = min(len(base), len(modified))
    for i in range(l):
        if base[i] != modified[i]:
            break
        head += 1
    if head == len(base) == len(modified):
        return head, 0
    for i in range(l):
        if base[-i-1] != modified[-i-1]:
            break
        tail += 1
    return head, tail


def test_head_and_tail():
    assert head_and_tail("abc", "abc") == (3, 0)
    assert head_and_tail("abc", "ab") == (2, 0)
    assert head_and_tail("abc", "b") == (0, 0)
    assert head_and_tail("abc", "c") == (0, 1)
    assert head_and_tail("abc", "xabc") == (0, 3)


DEBUG_UPLOAD_NOT_SEPARATE_PROCESS = False


class UploadProxy:
    def __init__(self):
        multiprocessing.set_start_method("spawn")
        self.upload_q = multiprocessing.Queue()
        self.cancelled_q = multiprocessing.Queue()
        if DEBUG_UPLOAD_NOT_SEPARATE_PROCESS:
            self.proc = None
        else:
            self.proc = multiprocessing.Process(
                target=_upload_results_loop,
                args=(self.upload_q, self.cancelled_q),
                name="upload_results",
                )
            self.proc.start()
        self._cancelled: Set[str] = set()

    def stop(self):
        if self.proc:
            self.upload_q.put(dict(exit=1))
            self.proc.join()
            self.proc = None

    def __del__(self):
        self.stop()

    def cancelled_reset(self):
        while not self.cancelled_q.empty():
            self._cancelled.add(self.cancelled_q.get())
        self._cancelled = set()

    def upload_result(
        self,
        description_dict: Dict[str, Any],
        original_batch: Dict[str, Any],
        *,
        status: str,                  # "in_progress", "completed"
        idx_updated: List[int],       # batch indexes where you have progress
        files: List[Dict[str, str]],  # updated text in those indexes
        finish_reason: List[str],     # empty if not finished yet
        tokens: Optional[List[int]] = None,
        more_toplevel_fields: Optional[List[Dict[str, Any]]] = None,
        generated_tokens_n: Optional[List[int]] = None,
        ts_arrived: float,
        ts_batch_started: float,
        ts_prompt: float,
        ts_first_token: float,
        ts_batch_finished: float,
    ):
        upload_dict = copy.deepcopy(description_dict)
        upload_dict["ts_arrived"] = ts_arrived
        upload_dict["ts_batch_started"] = ts_batch_started
        upload_dict["ts_prompt"] = ts_prompt
        upload_dict["ts_first_token"] = ts_first_token
        upload_dict["ts_batch_finished"] = ts_batch_finished
        progress = dict()
        for i, b in enumerate(idx_updated):
            progress[original_batch[b]["id"]] = {
                "id": original_batch[b]["id"],
                "stream": original_batch[b]["stream"],
                "object": "text_completion",
                "orig_files": original_batch[b]["sources"],
                "choices": [
                    {
                        "index": 0,
                        "files": files[i],
                        # "tokens": ([int(t) for t in tokens[b]] if tokens is not None else None),
                        "logprobs": None,
                        "finish_reason": finish_reason[i]
                    },
                ],
                "status": status,
                "created": original_batch[b]["created"],
                "more_toplevel_fields": (more_toplevel_fields[i] if more_toplevel_fields is not None else dict()),
                "generated_tokens_n": (generated_tokens_n[i] if generated_tokens_n is not None else 0),
            }
        upload_dict["progress"] = progress
        upload_dict["check_cancelled"] = [call["id"] for call in original_batch]
        upload_dict["model_name"] = description_dict["model"]
        self.upload_q.put(copy.deepcopy(upload_dict))
        if DEBUG_UPLOAD_NOT_SEPARATE_PROCESS:
            _upload_results_loop(self.upload_q, self.cancelled_q)

    def keepalive(self):
        self.upload_q.put(dict(keepalive=1))

    def check_cancelled(self):
        while not self.cancelled_q.empty():
            self._cancelled.add(self.cancelled_q.get())
        return self._cancelled


def _upload_results_loop(upload_q: multiprocessing.Queue, cancelled_q: multiprocessing.Queue):
    req_session = requests.Session()
    exit_flag = False
    while not exit_flag:
        try:
            upload_dict = upload_q.get(timeout=600)
        except queue.Empty as e:
            log("%s %s" % (datetime.datetime.now().strftime("%H:%M:%S.%f"), termcolor.colored("upload_results_loop timeout, exiting", "red")))
            exit_flag = True
            continue
        if "exit" in upload_dict:
            exit_flag = True
            break
        if "progress" not in upload_dict:
            continue
        t1 = time.time()
        while 1:
            if upload_dict.get("ts_batch_finished", 0) > 0:
                # Send ASAP
                break
            maybe_pile_up = upload_q.get() if not upload_q.empty() else None
            if maybe_pile_up is None:
                if time.time() < t1 + 0.1:
                    # Normally send every ~0.5 seconds
                    time.sleep(0.1)
                    continue
                else:
                    break
            if "exit" in maybe_pile_up:
                exit_flag = True
            if "progress" in maybe_pile_up:
                upload_dict["progress"].update(maybe_pile_up["progress"])
                upload_dict["ts_batch_finished"] = maybe_pile_up["ts_batch_finished"]
        resp = None
        # Remove head and tail if streaming, "files" becomes "files_head_mid_tail"
        for k, progress_dict in upload_dict["progress"].items():
            stream = progress_dict["stream"]
            orig_files = progress_dict.pop("orig_files")
            if not stream:
                continue
            stream_files = dict()
            for choice in progress_dict["choices"]:
                files = choice.pop("files")
                for k in files.keys():
                    orig = orig_files[k]
                    if not orig.endswith("\n"):
                        orig += "\n"
                    head, tail = head_and_tail(orig, files[k])
                    stream_files[k] = {
                        "head": head,
                        "mid": files[k][head:-tail],
                        "tail": tail,
                    }
                choice["files_head_mid_tail"] = stream_files
        t2 = time.time()
        for _attempt in range(5):
            j = dict()
            try:
                url = url_get_the_best() + "completion-upload-results"
                resp = req_session.post(url, json=upload_dict, timeout=2)
                j = resp.json()
            except requests.exceptions.ReadTimeout as e:
                t3 = time.time()
                log("%s %0.1fms %s %s" % (datetime.datetime.now().strftime("%H:%M:%S.%f"), 1000*(t3 - t2), url, termcolor.colored("TIMEOUT", "green")))
                url_complain_doesnt_work()
                continue
            except Exception as e:
                log("%s post response failed: %s\nServer response was: \"%s\"" % (url, str(e), resp.text[:150] if resp else "no response"))
                #if resp is not None:
                #    log("server response text:\n%s" % (resp.text,))
                url_complain_doesnt_work()
                continue
            if resp and resp.status_code != 200:
                log("%s post response failed: %i %s" % (url, resp.status_code, resp.text[:150]))
                url_complain_doesnt_work()
                continue
            break
        t3 = time.time()
        cancelled_n = 0
        if "cancelled" in j:
            for can in j["cancelled"]:
                cancelled_q.put(can)
                cancelled_n += 1
        log("%s %s %s %s %i uploaded, %i cancelled" % (datetime.datetime.now().strftime("%H:%M:%S.%f"),
            termcolor.colored("%0.1fms" % (1000*(t3 - t2),), "green"),
            url,
            j.get("retcode", "FAIL"),
            len(upload_dict["progress"]),
            cancelled_n,
            ))
        if j.get("retcode", "FAIL") != "OK":
            log("Server returned:", str(j))
        if DEBUG_UPLOAD_NOT_SEPARATE_PROCESS:
            break



import os, time, re, json, requests, termcolor, functools
from typing import Optional, List, Dict, Any, Union, Callable
import cloudpickle
from smallcloud import config, call_api, ssh_commands


MAX_UPLOAD_SIZE = 25*1024*1024


def upload_file(fn: str):
    config.read_config_file()
    size = os.path.getsize(fn)
    if size > MAX_UPLOAD_SIZE:
        raise Exception("Maximum archive size is %0.1fM, your code %0.1fM" % (MAX_UPLOAD_SIZE/1024/1024, size/1024/1024))
    url = config.v1_url + "task-file-upload"
    headers = config.account_and_secret_key_headers()
    files = {
        'file1': (os.path.basename(fn), open(fn, 'rb'), 'application/zip'),
    }
    print(url, "POST", os.path.basename(fn))
    r = requests.post(url, files=files, headers=headers)
    if r.status_code != 200:
        print("HTTP STATUS", r.status_code)
        quit(1)
    j = json.loads(r.text)
    if j["retcode"] != "OK":
        print(termcolor.colored(j["retcode"], "red"), j["human_readable_message"])
        quit(1)
    return j["upload_id"]


@functools.lru_cache()
def cached_code_to_zip():
    from smallcloud import code_to_zip
    return code_to_zip()


def launch_task(
    task_name: str,
    training_function: Union[Callable, str],
    args: List[Any] = [],
    kwargs: Dict[str, Any] = {},
    gpu_type="a5000",  # or more specifically cluster/gpu_type, "ant/a5000"
    gpus: Optional[int] = None,
    gpus_min: Optional[int] = None,
    gpus_max: Optional[int] = None,
    gpus_incr: int = 8,
    nice: int = 1,  # 0 preempts others, 1 normal, 2 low
    os_image: str = "",
    env: Dict[str, str] = {},
    pre_upload_code_zip: bool = False,  # Uploads code to cloud first (stored on our servers), otherwise uploads code after the task is started (good for privacy)
    hold_off_start: bool = False,       # Useful for debugging: set up everything, but don't start the task so you can start it under a debugger
    kill_upload_start: bool = False,
    force_node: str = "",
    call_function_directly: Optional[bool] = None,
):
    if call_function_directly or (call_function_directly is None and config.already_running_in_cloud):
        for k, v in env.items():
            os.environ[k] = v
        os.environ["TASK_NAME"] = task_name
        training_function(*args, **kwargs)
        return
    config.read_config_file()
    os.makedirs("/tmp/smc-temp", exist_ok=True)

    if not hold_off_start:
        pickle_filename = f"/tmp/smc-temp/pickle-call-{task_name}.pkl"
        cloudpickle.dump({
            "training_function": training_function,
            "args": args,
            "kwargs": kwargs,
            "env": {"TASK_NAME": task_name, **env},
            }, open(pickle_filename, "wb"))
    else:
        # Run the same script manually -- it will work without .pkl machinery, by calling function directly (see 'already_running_in_cloud' above).
        pickle_filename = ""

    post_json = {
        "task_name": task_name,
        "tenant_image": os_image,
        "gpu_type": gpu_type,
        "nice": nice,
        "force_node": force_node,
    }
    if gpus is not None:
        assert gpus_min is None and gpus_max is None
        post_json["gpus_min"] = gpus
        post_json["gpus_max"] = gpus
    else:
        assert gpus is None
        assert gpus_min is not None and gpus_max is not None, "Please set either gpus or gpus_min/gpus_max"
        post_json["gpus_min"] = gpus_min
        post_json["gpus_max"] = gpus_max
        post_json["gpus_incr"] = gpus_incr
    if pre_upload_code_zip:
        post_json["file_pkl"] = upload_file(pickle_filename) if pickle_filename else ""
        post_json["file_zip"] = upload_file(cached_code_to_zip())

    ret = call_api.fetch_json(config.v1_url + "reserve", post_json, headers=config.account_and_secret_key_headers(), ok_retcodes=["RUNNING"])
    call_api.pretty_print_response(ret)
    if ret["retcode"] == "RUNNING" and not kill_upload_start:
        print("Exiting. Set kill_upload_start=True to force a restart.")
        return
    kus = ret["retcode"] == "RUNNING"

    if pre_upload_code_zip and not kus:
        return
    if not kus:
        zip_filename = cached_code_to_zip()
    while 1:
        time.sleep(3)
        ret = call_api.fetch_json(
            config.v1_url + "task-nodes", get_params={"task_name": task_name},
            headers=config.account_and_secret_key_headers(),
            ok_retcodes=["WAIT"])
        if ret["retcode"] == "WAIT":
            print("%s waiting for the task to be scheduled" % time.strftime("%Y%m%d %H:%M:%S"))
            continue
        nodes = ret["nodes"]
        nodes_running = [(1 if x["status"]=="running" else 0) for x in nodes]
        status_list = [x["status"] for x in nodes]
        print("%s started %i/%i nodes %s" % (time.strftime("%Y%m%d %H:%M:%S"), sum(nodes_running), len(nodes), str(status_list)))
        if sum(nodes_running) == len(nodes) and len(nodes) > 0:
            break
    def waitall(ps, doing):
        ret = [p.wait() for p in ps]
        if any(ret):
            print("There was an error %s" % doing)
            quit(1)
    if len(nodes) > 1:
        ps = [ssh_commands.command_ssh(n["hostname"], "./smc_multinode_setup", fire_off=True) for n in nodes]
        waitall(ps, "setting up /etc/hosts and ssh keys for multi node")
    if not kus:
        ps = [ssh_commands.command_scp(zip_filename, n["hostname"] + ":code.7z", fire_off=True) for n in nodes[0:1]]
        waitall(ps, "copying code.7z to the first node")
    else:
        ps = [ssh_commands.command_ssh(n["hostname"], "killall python python3 || true", fire_off=True) for n in nodes]
        waitall(ps, "killing any existing python processes")
        ssh_commands.command_upload_code(nodes[0]["hostname"])
    ps = [ssh_commands.command_ssh(n["hostname"], "./smc_unpack_code.py", fire_off=True) for n in nodes[0:1]]
    waitall(ps, "running smc_unpack_code.py on the first node. Try looking at \"s tail %s\"" % nodes[0]["hostname"])
    if not hold_off_start:
        ps = [ssh_commands.command_scp(pickle_filename, n["hostname"] + ":pickled-function-call.pkl", fire_off=True) for n in nodes]
        waitall(ps, "copying pickled startup function call")
        # ps = [ssh_commands.command_ssh(n["hostname"], "nohup 2>&1 mpirun -n 8 -f mpihosts python smc_run_task.py | tee --append ~/output.log &", fire_off=True) for n in nodes]
        if len(nodes) > 1:
            ps = [ssh_commands.command_ssh(n["hostname"],
                'bash --login -c "nohup mpirun -f mpihosts ./smc_run_task.py >> ~/output.log 2>&1 &"',
                fire_off=True) for n in nodes[0:1]]
        else:
            ps = [ssh_commands.command_ssh(n["hostname"],
                'bash --login -c "nohup mpirun -n GPUS ./smc_run_task.py >> ~/output.log 2>&1 &"'.replace("GPUS", str(gpus or gpus_min)),
                fire_off=True) for n in nodes[0:1]]
        waitall(ps, "starting smc_run_task.py on the first node")


def shutdown():
    """
    Call this inside container to free up the node.
    """
    open("/tmp/shutdown-container", "a").close()



import json, time, termcolor
import urllib
import urllib.request
import urllib.error


def fetch_json(url, post_json=None, get_params=None, headers={}, ok_retcodes=[]):
    t0 = time.time()
    try:
        if get_params is not None:
            url += "?" + urllib.parse.urlencode(get_params)
        elif post_json is not None:
            print(json.dumps(post_json))
        req = urllib.request.Request(
            url,
            json.dumps(post_json).encode("utf-8") if post_json else None,
            {'Content-Type': 'application/json', **headers}
        )
        result = urllib.request.urlopen(req).read()
        t1 = time.time()
        print_if_appropriate("%0.2fs %s" % (t1 - t0, url))
    except urllib.error.URLError:
        print("ERROR %s" % (url))
        raise
    try:
        j = json.loads(result)
    except ValueError:
        print("Response from server is not a json:")
        print(result.decode("utf-8"))
        quit(1)
    if "retcode" in j and j["retcode"] not in (["OK"] + ok_retcodes):
        print(termcolor.colored(j["retcode"], "red"), j["human_readable_message"])
        quit(1)
    return j


def pretty_print_response(json):
    if isinstance(json, dict) and "retcode" in json:
        retcode = json["retcode"]
        color = "green" if retcode == "OK" else "red"
        print(termcolor.colored(retcode, color), json["human_readable_message"])
        return
    print(json)


global_option_json = False


def print_if_appropriate(*args):
    if not global_option_json:
        print(*args)


def print_table(resp, omit_for_brevity=[]):
    if global_option_json:
        print(json.dumps(resp, indent=4))
        return
    if len(resp) == 0:
        print("Empty result")
        return
    if isinstance(resp, dict):
        keys = sorted(resp.keys())
        assert isinstance(resp[keys[0]], dict)
        flatlist = [resp[k] for k in keys]
    elif isinstance(resp, list):
        flatlist = resp
    else:
        print("Server returned:\n%s" % str(resp))
        quit(1)
    def print_datetime(ts):
        if ts==0: return "-"
        full = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(ts))
        if full.startswith(time.strftime("%Y-%m-%d", time.localtime(time.time()))):
            return time.strftime("%H:%M:%S", time.localtime(ts))
        if full.startswith(time.strftime("%Y-%m-%d", time.localtime(time.time() - 86400))):
            return time.strftime("%a %H:%M:%S", time.localtime(ts))
        return full
    import pandas   # is slow, don't import at the top of the file.
    df = pandas.DataFrame()
    for column in flatlist[0].keys():
        if column in omit_for_brevity:
            continue
        if not column.startswith("ts_") and not column.endswith("_ts"):
            df[column.upper()] = [x[column] for x in flatlist]
        else:
            df[column.upper()] = [print_datetime(x[column]) for x in flatlist]
    pandas.set_option('display.max_rows', None)
    print(df)




import os, sys, requests, time, json
from typing import Union, Tuple, Generator, Optional


base_url = "https://inference.smallcloud.ai/v1/"


class APIConnectionError(Exception):
    pass


_dsess: Optional[requests.Session] = None


def default_session() -> requests.Session:
    global _dsess
    if _dsess is None:
        if "SMALLCLOUD_API_KEY" not in os.environ:
            raise ValueError("Please either set SMALLCLOUD_API_KEY environment variable or create requests session manually.")
        _dsess = requests.Session()
        _dsess.headers.update({
            "Authorization": "Bearer %s" % os.environ["SMALLCLOUD_API_KEY"],
        })
    return _dsess


def nlp_model_call(
    endpoint: str,
    model: str,
    *,
    req_session: Optional[requests.Session]=None,
    max_tokens: int,
    stream: bool=False,
    temperature: float,
    top_p: Optional[float]=None,
    top_n: Optional[int]=None,
    # TODO stop sequences
    verbose: int=0,
    **pass_args
) -> Union[Tuple[str, str], Generator[Tuple[str, str], None, None]]:
    req_session = req_session or default_session()
    assert isinstance(req_session, requests.Session)
    url = base_url + endpoint
    data = {
        "model": model,
        "temperature": temperature,
        "max_tokens": max_tokens,
        "stream": stream,
        **pass_args,
    }
    if top_p is not None:
        data["top_p"] = top_p
    if top_n is not None:
        data["top_n"] = top_n
    if verbose > 1:
        print("POST %s" % (data,))
    resp = None
    try:
        t0 = time.time()
        if not stream:
            resp = req_session.post(url, json=data)
        else:
            resp = req_session.post(url, json=data, stream=True)
        t1 = time.time()
        if verbose > 0:
            print("%0.1fms %s" % (1000*(t1 - t0), url))
    except Exception as e:
        if resp is not None:
            raise APIConnectionError("completions() json parse failed: %i\n%s" % (resp.status_code, resp.text))
        else:
            raise APIConnectionError("completions() failed: %s" % str(e))
    if resp.status_code != 200:
        raise APIConnectionError("status=%i, server returned:\n%s" % (resp.status_code, resp.text))
    if stream:
        def _streamer(resp):
            for line in resp.iter_lines(chunk_size=1):
                if line.startswith(b"data: {"):
                    j = json.loads(line[6:].decode("utf-8"))
                    yield j
                if line.startswith(b"data: [DONE]"):
                    return
        return _streamer(resp)
    if verbose > 1:
        print("RESPONSE", resp.text)
    try:
        json_resp = resp.json()
    except Exception as e:
        if resp is not None:
            raise APIConnectionError("completions() json parse failed: %i\n%s" % (resp.status_code, resp.text))
        else:
            raise APIConnectionError("completions() failed: %s" % str(e))
    return json_resp



from setuptools import setup

setup(
    name="smallcloud",
    py_modules=["smallcloud"],
    packages=["smallcloud"],
    version="0.3.1",
    url="https://github.com/smallcloudai/smallcloud",
    summary="Command line tool to access smallcloud.ai services",
    description="Run your GPU-intensive tasks using this command line tool / Python library",
    license='GNU GPLv3',
    install_requires=[
        "termcolor",
        "pandas",
    ],
    author="Small Magellanic Cloud AI Ltd.",
    author_email="cli-tool@smallcloud.tech",
    entry_points={
        "console_scripts": ["s = smallcloud.__main__:main"],
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Environment :: Console",
        "Operating System :: OS Independent",
    ]
)



import os, sys, subprocess, termcolor
from smallcloud import config, code_root, call_api


def run(cmd, stdout=None, stderr=None, colorize="", **kwargs):
    # This function runs 'rsync' and 'ssh-keygen'
    # To debug, use:
    #  verbose=1 dry=1 s command
    verbose = int(os.environ.get("verbose", "0"))
    if not verbose:
        stdout = subprocess.DEVNULL if stdout is None else stdout
    stderr = stderr or subprocess.PIPE
    if not call_api.global_option_json:
        print(" ".join(cmd) if not colorize else termcolor.colored(" ".join(cmd), colorize))
    dry = int(os.environ.get("dry", "0"))
    if dry:
        return 0
    completed_process = subprocess.run(cmd, stdout=stdout, stderr=stderr, **kwargs)
    if completed_process.returncode != 0:
        print("RETCODE: %s" % completed_process.returncode)
    if completed_process.stderr and (verbose or completed_process.returncode != 0):
        print("STDERR: %s" % completed_process.stderr.decode("utf-8"))
    return completed_process.returncode


def fetch_sshables():
    sshables = call_api.fetch_json(config.v1_url + "list-ssh-able", headers=config.account_and_secret_key_headers())
    known_hosts = []
    for rec in sshables:
        if rec["ed25519"].startswith("ssh-ed25519"):
            known_hosts.append("[%s]:%i %s" % (rec['ssh_addr'], rec['ssh_port'], rec['ed25519']))
    return sshables, known_hosts


def save_known_hosts(known_hosts):
    with open(config.known_hosts_file, "wt") as f:
        f.write("\n".join(known_hosts) + "\n")
    os.chmod(config.known_hosts_file, 0o600)


def add_ssh_identity_if_exists(ssh_cmdline):
    if os.path.exists(config.ssh_rsa_id_file):
        ssh_cmdline.extend(["-i", config.ssh_rsa_id_file])


def command_ssh(user_at_name, *args, fire_off=False):
    if "@" not in user_at_name:
        computer_name = user_at_name
        user = "user"
    else:
        user, computer_name = user_at_name.split("@")
    closest_match = None
    closest_match_dist = 1e10
    sshables, known_hosts = fetch_sshables()
    import difflib
    right_rec = None
    for rec in sshables:
        if rec["name"] == computer_name:
            if not rec["ed25519"]:
                print("'%s' is not ready" % rec["name"])
                return False
            right_rec = rec
        dist = difflib.SequenceMatcher(None, rec["name"], computer_name).ratio()
        if dist > 0.8 and dist < closest_match_dist:
            closest_match = rec
            closest_match_dist = dist
    if right_rec is None:
        call_api.print_table(sshables)
        print("Computer \"%s\" wasn't found." % computer_name)
        if closest_match is not None:
            print("Did you mean \"%s\"?" % closest_match["name"])
        return
    cmd = [
        "ssh",
        "%s@%s" % (user, right_rec['ssh_addr']),
        "-p", "%i" % right_rec['ssh_port'],
    ]
    if right_rec["ed25519"].startswith("ssh-ed25519"):  # Ether way strict checking is on!
        save_known_hosts(known_hosts)
        cmd.extend(["-o", "UserKnownHostsFile=%s" % config.known_hosts_file])
        add_ssh_identity_if_exists(cmd)
    cmd.extend(args)
    if fire_off:
        # return "a promise" client can wait on
        green_starts = -len(args)
        print(" ".join(cmd[:green_starts]) + " " + termcolor.colored(" ".join(cmd[green_starts:]), "green"))
        return subprocess.Popen(cmd)
    else:
        # this replaces the current process with scp
        print(" ".join(cmd))
        os.execv("/usr/bin/ssh", cmd)


def command_scp(*args, fire_off=False):
    remote_at = None
    for i in range(len(args)):
        if args[i].find(":") != -1:
            print("Re-writing \"%s\" as a remote location" % args[i])
            remote_at = i
            break
    if remote_at is None:
        print("Not clear which parameter refers to a remote location, this is detected by the presence of a colon \":\"")
        print("Examples:")
        print("s scp local_file1 job:remote_file")
        print("s scp \"user@job:remote_file*.txt\" local_folder/")
        quit(1)
    remote_location = args[remote_at]
    if "@" in remote_location:
        user, computer_name_colon_file = remote_location.split("@")
    else:
        user = "user"
        computer_name_colon_file = remote_location
    computer_name, path = computer_name_colon_file.split(":")
    right_rec = None
    sshables, known_hosts = fetch_sshables()
    for rec in sshables:
        if rec["name"] == computer_name:
            if not rec["ed25519"]:
                print("'%s' is not ready" % rec["name"])
                return False
            right_rec = rec
    if right_rec is None:
        call_api.print_table(sshables)
        print("Computer \"%s\" wasn't found." % computer_name)
        quit(1)
    cmd = ["scp", "-P", "%i" % right_rec['ssh_port']]
    if right_rec["ed25519"].startswith("ssh-ed25519"):
        save_known_hosts(known_hosts)
        cmd.extend(["-o", "UserKnownHostsFile=%s" % config.known_hosts_file])
        add_ssh_identity_if_exists(cmd)
    for i, a in enumerate(args):
        if i == remote_at:
            cmd.append("%s@%s:%s" % (user, right_rec['ssh_addr'], path))
        else:
            cmd.append(a)
    if fire_off:
        # return "a promise" client can wait on
        print(termcolor.colored(" ".join(cmd), "green"))
        return subprocess.Popen(cmd)
    else:
        # this replaces the current process with scp
        print(" ".join(cmd))
        os.execv("/usr/bin/scp", cmd)


def command_upload_code(*args):
    coderoot = code_root.detect_code_root()
    init = False
    if "--init" in args:
        init = True
        args = [x for x in args if x != "--init"]
    if len(args) == 0:
        print("Please specify computers to upload your code, for example \"myjob05*\", also try \"s list\".")
        quit(1)
    sshables, known_hosts = fetch_sshables()
    save_known_hosts(known_hosts)
    upload_dest = []
    upload_user = []
    for j in args:
        if "@" in j:
            user, computer_name = j.split("@")
        else:
            user = "user"
            computer_name = j
        for rec in sshables:
            import fnmatch
            if fnmatch.fnmatch(rec["name"], computer_name):
                if not rec["ed25519"]:
                    print("'%s' is not ready" % rec["name"])
                    return False
                upload_dest.append(rec)
                upload_user.append(user)
    call_api.print_if_appropriate("Uploading code to:")
    call_api.print_table(upload_dest, omit_for_brevity="ed25519")
    for rec, user in zip(upload_dest, upload_user):
        # "-u" update based on modification time
        # "-c" update based on checksum, not date, because git might clone newer files than your modified ones
        # "--delete" -- nice to have, but has unexpected effects
        ssh_cmd = [
            "ssh",
            "-p", "%i" % rec["ssh_port"],
        ]
        if rec["ed25519"].startswith("ssh-ed25519"):
            add_ssh_identity_if_exists(ssh_cmd)
            ssh_cmd.extend(["-o", "UserKnownHostsFile=%s" % config.known_hosts_file])
        cmd = [
            "rsync", "-rpl", "-c", "--itemize-changes", coderoot, f"{user}@{rec['ssh_addr']}:code/", "--filter=:- .gitignore", "--exclude=.git",
            "-e", " ".join(ssh_cmd),
            ]
        r = run(cmd, stdout=sys.stdout, stderr=sys.stderr, colorize="green")
        assert r==0, r
        if init:
            ssh_cmd.extend([f"{user}@{rec['ssh_addr']}", "cd code && bash .smc_code_root.sh"])
            r = run(ssh_cmd, stdout=sys.stdout, stderr=sys.stderr, colorize="green")
            assert r==0, r
    return True


def command_ssh_keygen(*args):
    jobs_for_warning = call_api.fetch_json(config.v1_url + "jobs", headers=config.account_and_secret_key_headers())
    jobs_running = [x for x in jobs_for_warning if x["ts_finished"] == 0]
    if len(jobs_running) > 0:
        print(f"You have {len(jobs_running)} jobs running. All ssh-based commands from this computer will start to use a new \"-i {config.ssh_rsa_id_file}\" identity file, this might prevent you from logging in to these running machines.")
        quit(1)
    try:
        os.unlink(config.ssh_rsa_id_file)
    except FileNotFoundError:
        pass
    r = run(["ssh-keygen", "-f", config.ssh_rsa_id_file, "-N", "", *args])
    assert r==0, r
    resp = call_api.fetch_json(
        config.v1_url + "ssh-public-key-upload",
        post_json={"ssh_public_key": open(config.ssh_rsa_id_file + ".pub").read()},
        headers=config.account_and_secret_key_headers())
    call_api.pretty_print_response(resp)


def command_ssh_key_upload(*args):
    if len(args) != 1:
        print("Please specify a file to upload, such as ~/.ssh/id_rsa.pub\n(do this if you want ssh without -i option to work, for a dedicated key use \"s ssh-keygen\")")
        quit(1)
    resp = call_api.fetch_json(
        config.v1_url + "ssh-public-key-upload",
        post_json={"ssh_public_key": open(os.path.expanduser(args[0])).read()},
        headers=config.account_and_secret_key_headers())
    call_api.pretty_print_response(resp)
